# -*- coding: utf-8 -*-
from setuphelpers import *
from setupdevhelpers import *
import waptguihelper
import uuid
import json
import requests
import traceback

# Products configuration
# Latest ID can be found official microsoft website or here : https://msdl.gravesoft.dev/ 
PRODUCTS = {
    # "48": "Windows 8.1 Single Language (9600.17415)",
    # "52": "Windows 8.1 (9600.17415)",
    # "55": "Windows 8.1 N (9600.17415)",
    # "61": "Windows 8.1 K (9600.17415)",
    # "62": "Windows 8.1 KN (9600.17415)",
    # "2378": "Windows 10 22H2 Home China (19045.2006)",
    # "2618": "Windows 10 22H2 v1 (19045.2965)",
    "3113": "Windows 11 24H2 (26100.1742)",
    # "3114": "Windows 11 24H2 Home China (26100.1742)",
    # "3115": "Windows 11 24H2 Pro China (26100.1742)",
    # "3131": "Windows 11 Arm64 24H2 (26100.1742)",
    # "3132": "Windows 11 Arm64 24H2 Home China (26100.1742)",
    # "3133": "Windows 11 Arm64 24H2 Pro China (26100.1742)"
}


class WindowsDownloader:
    def __init__(self, debug=False, fallback_msdl=True, proxies=None):
        self.debug = debug
        self.fallback_msdl = fallback_msdl
        self.proxies = proxies
        self.session = self._initialize_session()
        self.session_id = str(uuid.uuid4())
        self.SESSION_URL = "https://vlscppe.microsoft.com/tags?org_id=y6jn8c31&session_id="
        self.MS_API_URL = "https://www.microsoft.com/software-download-connector/api"
        self.MSDL_API_URL = "https://api.gravesoft.dev/msdl"
        self.PARAMS = "profile=606624d44113&Locale=en-US&sessionID="
        self.SHARED_SESSION_GUID = "47cbc254-4a79-4be6-9866-9c625eb20911"

        # Needed to validate links retrieved by MSDL proxy
        self.MICROSOFT_DOWNLOAD_URL = "https://software.download.prss.microsoft.com"

        self.prepare_download(self.session_id)

    def _initialize_session(self):
        session = requests.Session()
        if self.proxies :
            session.proxies.update(self.proxies)
        session.headers.update({
            "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:128.0) Gecko/20100101 Firefox/128.0",
            "Accept": "application/json, text/javascript, */*; q=0.01",
            "x-requested-with": "XMLHttpRequest",
            "sec-ch-ua-platform": "Windows",
            "cache-control": "no-cache"
        })
        return session

    def log_debug(self, message):
        if self.debug:
            print(f"[DEBUG] {message}")

    def prepare_download(self, session_id):
        url = f"{self.SESSION_URL}{session_id}"
        self.log_debug(f"Preparing session with URL: {url}")
        try:
            response = self.session.get(url)
            response.raise_for_status()
        except requests.RequestException as e:
            # print(traceback.format_exc())
            print(f"Error preparing session: {e}")

    def fetch_data(self, url):
        self.log_debug(f"Fetching data from URL: {url}")
        try:
            response = self.session.get(url)
            response.raise_for_status()
            result = response.json()
            if isinstance(result, str):
                return json.loads(result)
            return response.json()
        except requests.RequestException as e:
            # print(traceback.format_exc())
            print(f"Error preparing session: {e}")
            return None

    def get_languages(self, product_id):
        url = f"{self.MS_API_URL}/getskuinformationbyproductedition?{self.PARAMS}{self.session_id}&ProductEditionId={product_id}"
        return self.fetch_data(url)

    def get_download_links(self, product_id, sku_id):
        # Attempt to get download links from MS API
        downloads = self._get_ms_download_links(product_id, sku_id)
        if downloads:
            return downloads

        # Retry with shared session GUID only if the first request not send False
        # That mean we have been banned from microsoft
        downloads = self._retry_ms_download_links(product_id, sku_id) if downloads is not False else None
        if downloads:
            return downloads

        # Fallback to MSDL API if enabled
        if self.fallback_msdl:
            downloads = self._get_download_links_msdl(product_id, sku_id)
            return downloads

        print("[ERROR] No download links available after all attempts.")
        return None

    def _get_ms_download_links(self, product_id, sku_id, session_guid=None):
        session_guid = session_guid or self.session_id
        url = f"{self.MS_API_URL}/GetProductDownloadLinksBySku?{self.PARAMS}{session_guid}&SKU={sku_id}&friendlyFileName=undefined&ProductEditionId=undefined"

        product_name = PRODUCTS[product_id]
        if "Windows 8.1" in product_name:
            windows_version = "8.1"
        elif "Windows 10" in product_name:
            windows_version = "10"
        elif "Windows 11" in product_name:
            windows_version = "11"

        self.session.headers.update({
            "referer": f"https://www.microsoft.com/fr-fr/software-download/windows{windows_version}"
        })
        result = self.fetch_data(url)
        if result and result.get('Errors'):
            self.log_debug(result["Errors"])
            if result.get('Errors')[0]['Type'] == 9:
                print("[ERROR] No download links available, You've been potentialy banned !")
                return False
            print("[WARNING] Errors found in microsoft download link response. (Trying another method)")
            return None
        return result

    def _retry_ms_download_links(self, product_id, sku_id):
        self.session.cookies.clear()
        self.prepare_download(self.SHARED_SESSION_GUID)
        return self._get_ms_download_links(product_id, sku_id, session_guid=self.SHARED_SESSION_GUID)

    def _get_download_links_msdl(self, product_id, sku_id):
        url = f"{self.MSDL_API_URL}/proxy?product_id={product_id}&sku_id={sku_id}"

        self.session.cookies.clear()
        del self.session.headers['referer']

        downloads = self.fetch_data(url)

        self.log_debug('Validating MSDL download links.')
        if all(self.MICROSOFT_DOWNLOAD_URL in option.get('Uri', '') for option in downloads.get("ProductDownloadOptions", [])):
            return downloads

        return None

def update_package():
    # Declaring local variables
    package_updated = False
    proxies = get_proxies()
    if not proxies:
        proxies = get_proxies_from_wapt_console()

    iso_path = makepath(basedir, "iso")
    setup_path = makepath(iso_path, "setup.exe")
    iso_present = False
    iso_extracted = False
    iso_copy_path = None
    iso_to_download = False
    keep_extracted_iso = False
    keep_current_iso = False

    lang_dict = {
        "fr": "French",
        "en": "English",
        "pl": "Polish",
        "de": "German",
        "es": "Spanish",
        "pt": "Portuguese",
        "it": "Italian",
        "nl": "Dutch",
        "ru": "Russian",
        "all": "Others",
    }
    lang_list = []
    for entry in lang_dict:
        lang_list.append({"short name": entry, "long name": lang_dict[entry]})

    # Detect ISO if placed
    for iso_in_dir in glob.glob("*.iso"):
        if isfile(iso_in_dir):
            iso_present = True

    # Detect ISO if already extracted
    if isfile(setup_path):
        iso_extracted = True

    if iso_present:
        keep_iso_present_question = waptguihelper.message_dialog("ISO found", f"Do you want to keep: {iso_in_dir}?", waptguihelper.ID_YES)
        if keep_iso_present_question == waptguihelper.ID_YES:
            keep_current_iso = True

    if iso_extracted:
        keep_iso_extracted_question = waptguihelper.message_dialog("An ISO is already extracted", "Do you want to keep it?", waptguihelper.ID_YES)
        if keep_iso_extracted_question == waptguihelper.ID_YES:
            keep_extracted_iso = True

    if keep_current_iso and keep_extracted_iso:
        keep_only_one_question = waptguihelper.grid_dialog(
            "Please select only one", [{"keep :": iso_in_dir}, {"keep :": "iso already extracted"}], waptguihelper.GRT_SELECTED
        )
        if "extracted" in keep_only_one_question[0]["keep :"]:
            keep_current_iso = False
        else:
            keep_extracted_iso = False

    if not keep_extracted_iso:
        if not iso_present:
            iso_absent_question = waptguihelper.message_dialog(
                "No ISO detected", "Do you want to download Windows ISO from Microsoft's servers?\nClick No to use your own ISO", waptguihelper.ID_YES
            )
            if iso_absent_question == waptguihelper.ID_YES:
                iso_to_download = True

    if iso_to_download:
        downloader = WindowsDownloader(fallback_msdl=True, proxies=proxies)

        products_waptgui = [ {'id': id, 'name': name} for id, name in PRODUCTS.items() ]
        product_id = waptguihelper.grid_dialog("Please select a product", products_waptgui, waptguihelper.GRT_SELECTED)[0]['id']

        languages = downloader.get_languages(product_id)
        if not languages:
            error('Unable to retrieved languages from microsoft')
        
        languages_waptgui = [ {"id": sku['Id'], 'language': sku['LocalizedLanguage']} for sku in languages.get("Skus", []) ]
        sku_id = waptguihelper.grid_dialog("Please select language", languages_waptgui, waptguihelper.GRT_SELECTED)[0]['id']

        downloads = downloader.get_download_links(product_id, sku_id)
        if downloads:
            iso_links = downloads.get('ProductDownloadOptions', [])
            if len(iso_links) > 1:
                iso_name_waptgui = [ {"name": iso['Uri'].rsplit('/', 1)[-1].split('?')[0]} for iso in iso_links ]
                iso_name = waptguihelper.grid_dialog(
                    "Please select an ISO", iso_name_waptgui, waptguihelper.GRT_SELECTED, 
                    '{"columns":[{"propertyname":"name","width":375}],"window":{"height":331,"width":380}}'
                )[0]['name']
                download_url = next((iso['Uri'] for iso in iso_links if iso_name in iso['Uri']), None)
            else:
                download_url = downloads['ProductDownloadOptions'][0]['Uri']
                iso_name = download_url.rsplit('/', 1)[-1].split('?')[0]
                
            print(f"Downloading {iso_name}...")
            wget(download_url, makepath(basedir, iso_name), proxies=proxies)
            iso_present = True
        else:
            error("No download links available")

    if not keep_current_iso and not iso_present and not keep_extracted_iso:
        # Cleanup ISO files
        for iso_in_dir in glob.glob("*.iso"):
            remove_file(iso_in_dir)

        iso_copy_path = waptguihelper.filename_dialog("Please provide the Windows ISO to copy in this package", "", "", "ISO Files|*.iso")
        print("Copying: " + iso_copy_path)
        filecopyto(iso_copy_path, basedir)

    # Extracting ISO
    if not keep_extracted_iso:
        # Cleanup ISO dir
        if isdir(iso_path):
            remove_tree(iso_path)
        iso_in_dir = glob.glob("*.iso")[0]
        print("Extracting: " + iso_in_dir)
        unzip_with_7zip(iso_in_dir, iso_path)
        remove_file(iso_in_dir)

    # Getting version from the ISO setup.exe
    version = ".".join(get_file_properties(setup_path)["ProductVersion"].split(".")[:3])

    # Changing version of the package
    if Version(version) > control.get_software_version():
        print("Software version updated from: %s to: %s" % (control.get_software_version(), Version(version)))
        package_updated = True
    control.set_software_version(version)
    control.save_control_to_wapt()

    # Changing locale of the package
    lang_short = waptguihelper.grid_dialog("Please select a language", lang_list, waptguihelper.GRT_SELECTED)[0]["short name"]

    # Updating pkg infos
    if "-template" in control.package:
        package_name = waptguihelper.input_dialog(control.package, "You can redefine the package name", control.package.split("-template")[0])
        control.package = package_name
        if " Template" in control.name:
            control.name = control.name.split(" Template")[0]
        print("Changing package name to: %s" % package_name)
    control.locale = lang_short
    control.save_control_to_wapt()
    print("Changing locale to: %s in WAPT\\control" % control.locale)

    # Validating update-package-sources
    return package_updated
